from unittest.mock import MagicMock, patch

import pytest
from takeoff_client import TakeoffClient, TakeoffError
import responses


class TestTakeoffClient:
    def test_initialization_default(self):
        client = TakeoffClient()
        assert client.base_url == "http://localhost"
        assert client.port == 3000
        assert client.mgmt_port == 3001
        assert client.url == "http://localhost:3000"
        assert client.mgmt_url == "http://localhost:3001"

    def test_initialization_custom(self):
        # test that mgmt_port is set to port + 1
        client = TakeoffClient("http://test", 9000)
        assert client.base_url == "http://test"
        assert client.port == 9000
        assert client.mgmt_port == 9001
        assert client.url == "http://test:9000"
        assert client.mgmt_url == "http://test:9001"

    def test_initalization_custom_mgmt_port(self):
        client = TakeoffClient("http://test", 9000, 9005)
        assert client.base_url == "http://test"
        assert client.port == 9000
        assert client.mgmt_port == 9005
        assert client.url == "http://test:9000"
        assert client.mgmt_url == "http://test:9005"

    def test_get_readers(self):
        mock_response = {
            "primary": [
                {
                    "reader_id": "test",
                    "backend": "test",
                    "model_name": "test",
                    "model_type": "test",
                    "pids": [0],
                    "ready": True,
                }
            ]
        }

        with patch("requests.get") as mock_get:
            mock_get.return_value.json.return_value = mock_response

            client = TakeoffClient()
            response = client.get_readers()

            mock_get.assert_called_once_with("http://localhost:3001/reader_groups")
            assert response == mock_response

    def test_embed(self):
        mock_response = {"text": ["embedded text"]}

        with patch("requests.post") as mock_post:
            mock_post.return_value.json.return_value = mock_response
            mock_post.return_value.status_code = 200

            client = TakeoffClient()
            response = client.embed("text to embed")

            mock_post.assert_called_once()
            assert response == mock_response

            # Test that an Exception is raised when the status code is not 200
            mock_post.return_value.status_code = 400
            mock_post.return_value.text = "Error message"

            with pytest.raises(Exception) as e:
                client.embed("text to embed")

            assert str(e.value) == "Embedding failed\nStatus code: 400\nResponse: Error message"

    def test_generate(self):
        mock_response = {"text": "generated text"}

        with patch("requests.post") as mock_post:
            mock_post.return_value.json.return_value = mock_response
            mock_post.return_value.status_code = 200

            client = TakeoffClient()
            response = client.generate("text to generate")

            mock_post.assert_called_once()
            assert response == mock_response

            # Test that an Exception is raised when the status code is not 200
            mock_post.return_value.status_code = 400
            mock_post.return_value.text = "Error message"

            with pytest.raises(Exception) as e:
                client.generate("text to generate")

            assert str(e.value) == "Generation failed\nStatus code: 400\nResponse: Error message"

    @pytest.fixture
    def mock_response(self):
        # Mocking a requests.Response object
        mock_response = MagicMock()
        mock_response.status_code = 200

        def printer(x):
            print(x)
            return x

        alist = [b"data: Generated text 1\n\n", b"data: Generated text 2\n\n"]

        mock_response.__iter__.return_value = (printer(x) for x in alist)
        return mock_response

    def test_generate_stream_success(self, mock_response):
        # Setup
        text = "test text"
        url = "http://example.com/generate_stream"

        # Patching requests.post to return the mock_response
        with patch("requests.post", return_value=mock_response) as mock_post:
            instance = TakeoffClient()
            instance.url = "http://example.com"

            # Action
            result_generator = instance.generate_stream(text)
            result = list(result_generator)

            # Assert
            mock_post.assert_called_once_with(
                url=url,
                json={
                    "text": text,
                    "sampling_temperature": None,
                    "sampling_topp": None,
                    "sampling_topk": None,
                    "repetition_penalty": None,
                    "no_repeat_ngram_size": None,
                    "max_new_tokens": None,
                    "min_new_tokens": None,
                    "regex_string": None,
                    "json_schema": None,
                    "prompt_max_tokens": None,
                    "consumer_group": "primary",
                },
                stream=True,
            )

            print(result)
            assert len(result) == 2
            assert result[0].data == "Generated text 1"
            assert result[1].data == "Generated text 2"

    def test_generate_stream_failure(self, mock_response):
        # Setup for failure scenario
        mock_response.status_code = 400
        mock_response.text = "Error message"
        text = "test text"

        with patch("requests.post", return_value=mock_response):
            instance = TakeoffClient()
            instance.url = "http://example.com"

            # Assert exception is raised on failure
            with pytest.raises(Exception) as excinfo:
                _ = list(instance.generate_stream(text))

            assert "Generation failed" in str(excinfo.value)


    @responses.activate
    def test_create_reader(self):
        responses.add(
            responses.POST,
            "http://localhost:3001/reader",
            json={"reader_id": "test"},
            status=201,
        )

        client = TakeoffClient()
        response = client.create_reader({"model_name": "test", "device": "cuda"})

        assert response == {"reader_id": "test"}
    
    @responses.activate
    def test_reader_create_failure(self):
        responses.add(
            responses.POST,
            "http://localhost:3001/reader",
            status=400,
        )

        client = TakeoffClient()
        with pytest.raises(TakeoffError) as e:
            client.create_reader({"model_name": "test", "device": "cuda"})

        assert str(e.value) == "Reader creation failed\nStatus code: 400\nResponse: "
        
    @responses.activate
    def test_reader_create_failure_bad_device(self):
        client = TakeoffClient()
        # No device specified
        with pytest.raises(TakeoffError) as e:
            client.create_reader({"model_name": "test", "device": "mars"})
        assert "device" in str(e.value)
        # Wrong device
        with pytest.raises(TakeoffError) as e:
            client.create_reader({"model_name": "test"})
        assert "device" in str(e.value)

    @responses.activate
    def test_reader_create_failure_bad_model_name(self):
        client = TakeoffClient()
        # No model specified
        with pytest.raises(TakeoffError) as e:
            client.create_reader({"device": "mars"})
        assert "model_name" in str(e.value)
        # Empty model string
        with pytest.raises(TakeoffError) as e:
            client.create_reader({"model_name": ""})
        assert "model_name" in str(e.value)
        
    @responses.activate
    def test_reader_delete(self):
        responses.add(
            responses.DELETE,
            "http://localhost:3001/reader/test",
            status=204,
        )

        client = TakeoffClient()
        response = client.delete_reader("test")

        assert response == None
    
    @responses.activate
    def test_reader_delete_failure(self):
        responses.add(
            responses.DELETE,
            "http://localhost:3001/reader/test",
            status=404,
        )

        client = TakeoffClient()
        with pytest.raises(TakeoffError) as e:
            client.delete_reader("test")

        assert str(e.value) == "Reader deletion failed\nStatus code: 404\nResponse: "
        
    @responses.activate
    def test_list_reader_groups(self):
        responses.add(
            responses.GET,
            "http://localhost:3001/reader_groups",
            json={"primary": {"reader_id": "test"}},
            status=200,
        )

        client = TakeoffClient()
        response = client.list_all_readers()

        assert response == {"primary": {"reader_id": "test"}}
    
    @responses.activate
    def test_list_reader_groups_failure(self):
        responses.add(
            responses.GET,
            "http://localhost:3001/reader_groups",
            status=404,
        )

        client = TakeoffClient()
        with pytest.raises(TakeoffError) as e:
            client.list_all_readers()

        assert str(e.value) == "Reader listing failed\nStatus code: 404\nResponse: "
    
    @responses.activate
    def test_get_reader_config(self):
        responses.add(
            responses.GET,
            "http://localhost:3001/config/test",
            json={"reader_id": "test"},
            status=200,
        )

        client = TakeoffClient()
        response = client.get_reader_config("test")

        assert response == {"reader_id": "test"}
        
    @responses.activate
    def test_get_reader_config_failure(self):
        responses.add(
            responses.GET,
            "http://localhost:3001/config/test",
            status=404,
        )

        client = TakeoffClient()
        with pytest.raises(TakeoffError) as e:
            client.get_reader_config("test")

        assert str(e.value) == "Reader config retrieval failed\nStatus code: 404\nResponse: "


"""This module contains the TakeoffClient class, which is used to interact with the Takeoff server."""

# ────────────────────────────────────────────────────── Import ────────────────────────────────────────────────────── #
import json
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional, Union

import requests
from sseclient import Event, SSEClient
from takeoff_config import ReaderConfig

from .exceptions import TakeoffError

# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #
#                                                    Takeoff Client                                                    #
# ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────── #


class TakeoffClient:
    def __init__(self, base_url: str = "http://localhost", port: int = 3000, mgmt_port: int = None):
        """TakeoffClient is used to interact with the Takeoff server.

        Args:
            base_url (str, optional): base url that takeoff server runs on. Defaults to "http://localhost".
            port (int, optional): port that main server runs on. Defaults to 8000.
            mgmt_port (int, optional): port that management api runs on. Usually be `port + 1`. Defaults to None.
        """
        self.base_url = base_url  # "http://localhost"
        self.port = port  # 8000

        if mgmt_port is None:
            self.mgmt_port = port + 1  # 8001
        else:
            self.mgmt_port = mgmt_port

        self.url = f"{self.base_url}:{self.port}"  # "http://localhost:3000"
        self.mgmt_url = f"{self.base_url}:{self.mgmt_port}"  # "http://localhost:3001"

    def get_readers(self) -> dict:
        """Get a list of information about all readers.

        Returns:
            dict: List of information about all readers.
        """
        response = requests.get(self.mgmt_url + "/reader_groups")
        return response.json()

    def embed(self, text: Union[str, List[str]], consumer_group: str = "primary") -> dict:
        """Embed a batch of text.

        Args:
            text (str | List[str]): Text to embed.
            consumer_group (str, optional): consumer group to use. Defaults to "primary".

        Returns:
            dict: Embedding response.
        """
        response = requests.post(self.url + "/embed", json={"text": text, "consumer_group": consumer_group})

        if response.status_code != 200:
            raise TakeoffError(
                status_code=response.status_code,
                message=f"Embedding failed\nStatus code: {str(response.status_code)}\nResponse: {response.text}",
            )

        return response.json()

    def classify(self, text: Union[str, List[str], List[List[str]]], consumer_group: str = "classify") -> dict:
        """Classify a batch of text.

        Text that is passed in as a list of list of strings will be concatenated on the
        innermost list, and the outermost list treated as a batch of concatenated strings.

        Concatenation happens server-side, as it needs information from the model tokenizer.

        Args:
            text (str | List[str] | List[List[str]]): Text to classify.
            consumer_group (str, optional): consumer group to use. Defaults to "classify".

        Returns:
            dict: Classification response.
        """

        response = requests.post(self.url + "/classify", json={"text": text, "consumer_group": consumer_group})
        if response.status_code != 200:
            raise Exception(
                f"Classification failed\nStatus code: {str(response.status_code)}\nResponse: {response.text}"
            )

        return response.json()

    def generate(
        self,
        text: Union[str, List[str]],
        sampling_temperature: float = None,
        sampling_topp: float = None,
        sampling_topk: int = None,
        repetition_penalty: float = None,
        no_repeat_ngram_size: int = None,
        max_new_tokens: int = None,
        min_new_tokens: int = None,
        regex_string: str = None,
        json_schema: Any = None,
        prompt_max_tokens: int = None,
        consumer_group: str = "primary",
        image_path: Optional[Path] = None,
    ) -> dict:
        """Generates text, seeking a completion for the input prompt. Buffers output and returns at once.

        Args:
            text (str): Input prompt from which to generate
            sampling_topp (float, optional): Sample from set of tokens whose cumulative probability exceeds this value
            sampling_temperature (float, optional): Sample predictions from the top K most probable candidates
            sampling_topk (int, optional): Sample with randomness. Bigger temperatures are associated with more randomness.
            repetition_penalty (float, optional): Penalise the generation of tokens that have been generated before. Set to > 1 to penalize.
            no_repeat_ngram_size (int, optional): Prevent repetitions of ngrams of this size.
            max_new_tokens (int, optional): The maximum number of (new) tokens that the model will generate.
            min_new_tokens (int, optional): The minimum number of (new) tokens that the model will generate.
            regex_string (str, optional): The regex string which generations will adhere to as they decode.
            json_schema (dict, optional): The JSON Schema which generations will adhere to as they decode. Ignored if regex_str is set.
            prompt_max_tokens (int, optional): The maximum length (in tokens) for this prompt. Prompts longer than this value will be truncated.
            consumer_group (str, optional): The consumer group to which to send the request.
            image_path (Path, optional): Path to the image file to be used as input. Defaults to None.
                                         Note: This is only available if the running model supports image to text generation, for
                                         example with LlaVa models.

        Returns:
            Output (dict): The response from Takeoff containing the generated text as a whole.

        """
        json_data = {
            "text": text,
            "sampling_temperature": sampling_temperature,
            "sampling_topp": sampling_topp,
            "sampling_topk": sampling_topk,
            "repetition_penalty": repetition_penalty,
            "no_repeat_ngram_size": no_repeat_ngram_size,
            "max_new_tokens": max_new_tokens,
            "min_new_tokens": min_new_tokens,
            "regex_string": regex_string,
            "json_schema": json_schema,
            "prompt_max_tokens": prompt_max_tokens,
            "consumer_group": consumer_group,
        }

        if image_path is not None:
            with open(image_path, "rb") as f:
                # The payload needs to be converted to a string and then wrapped in a tuple with the content type
                files = {
                    # The 'json_data' is part of the multipart form data, with a content type of application/json
                    "json_data": (None, json.dumps(json_data), "application/json"),
                    # The image is sent as a file with the key 'image_data'
                    "image_data": (None, f, "image/*"),
                }

                response = requests.post(self.url + "/image_generate", files=files)
        else:
            response = requests.post(
                url=self.url + "/generate",
                json=json_data,
                stream=True,
            )

        if response.status_code != 200:
            raise TakeoffError(
                status_code=response.status_code,
                message=f"Generation failed\nStatus code: {str(response.status_code)}\nResponse: {response.text}",
            )

        return response.json()

    def generate_stream(
        self,
        text: Union[str, List[str]],
        sampling_temperature: float = None,
        sampling_topp: float = None,
        sampling_topk: int = None,
        repetition_penalty: float = None,
        no_repeat_ngram_size: int = None,
        max_new_tokens: int = None,
        min_new_tokens: int = None,
        regex_string: str = None,
        json_schema: dict = None,
        prompt_max_tokens: int = None,
        consumer_group: str = "primary",
        image_path: Optional[Path] = None,
    ) -> Iterator[Event]:
        """Generates text, seeking a completion for the input prompt.

        Args:
            text (str | List[str]): Input prompt from which to generate
            sampling_temperature (float, optional): Sample predictions from the top K most probable candidates
            sampling_topp (float, optional): Sample from set of tokens whose cumulative probability exceeds this value
            sampling_topk (int, optional): Sample with randomness. Bigger temperatures are associated with more randomness.
            repetition_penalty (float, optional): Penalise the generation of tokens that have been generated before. Set to > 1 to penalize.
            no_repeat_ngram_size (int, optional): Prevent repetitions of ngrams of this size.
            max_new_tokens (int, optional): The maximum number of (new) tokens that the model will generate.
            min_new_tokens (int, optional): The minimum number of (new) tokens that the model will generate.
            regex_string (str, optional): The regex string which generations will adhere to as they decode.
            json_schema (dict, optional): The JSON Schema which generations will adhere to as they decode. Ignored if regex_str is set.
            prompt_max_tokens (int, optional): The maximum length (in tokens) for this prompt. Prompts longer than this value will be truncated.
            consumer_group (str, optional): The consumer group to which to send the request.
            image_path (Path, optional): Path to the image file to be used as input. Defaults to None.
                                         Note: This is only available if the running model supports image to text generation, for
                                         example with LlaVa models.


        Returns:
            Iterator[sseclient.SSEClient.Event]: An iterator of server-sent events.

        """
        json_data = {
            "text": text,
            "sampling_temperature": sampling_temperature,
            "sampling_topp": sampling_topp,
            "sampling_topk": sampling_topk,
            "repetition_penalty": repetition_penalty,
            "no_repeat_ngram_size": no_repeat_ngram_size,
            "max_new_tokens": max_new_tokens,
            "min_new_tokens": min_new_tokens,
            "regex_string": regex_string,
            "json_schema": json_schema,
            "prompt_max_tokens": prompt_max_tokens,
            "consumer_group": consumer_group,
        }
        if image_path is not None:
            with open(image_path, "rb") as f:
                # The payload needs to be converted to a string and then wrapped in a tuple with the content type
                files = {
                    # The 'json_data' is part of the multipart form data, with a content type of application/json
                    "json_data": (None, json.dumps(json_data), "application/json"),
                    # The image is sent as a file with the key 'image_data'
                    "image_data": (None, f, "image/*"),
                }

                response = requests.post(self.url + "/image_generate", files=files)
        else:
            response = requests.post(
                url=self.url + "/generate_stream",
                json=json_data,
                stream=True,
            )
        if response.status_code != 200:
            raise TakeoffError(
                status_code=response.status_code,
                message=f"Generation failed\nStatus code: {str(response.status_code)}\nResponse: {response.text}",
            )

        # return a python generator
        return SSEClient(response).events()

    def tokenize(self, text: str, reader_id: str) -> List[str]:
        """Tokenize a single text item.

        The tokenize endpoint can be used to send a string to a models tokenizer for tokenization.
        The result is a list of tokens. For example, if "my_reader" is the id of a model that uses a Llama tokenizer,
        The following code will tokenize the string "hello, world" using the Llama tokenizer:

        >>> takeoff_client.tokenize("hello, world", reader_id="my_reader")
        ... ['▁hello', ',', '▁world']

        NOTE: The `reader_id` parameter is not the same as the `consumer_group` parameter used in other endpoints.
        Because tokenization is specific to a specific loaded model, we need to specify a unique id that identifies
        a particular reader. To find this ID for the models currently loaded into your takeoff server, try the following

        >>> readers = takeoff_client.get_readers()
        >>> for reader_group in readers.values():
        >>>    for reader in reader_group:
        >>>       print(reader["reader_id"])

        Args:
            text (str): Text to tokenize.
            reader_id (str): The id of the reader to use.

        Returns:
            List[str]: Tokenized text.
        """
        response = requests.post(self.url + f"/tokenize/{reader_id}", json={"text": text})
        if response.status_code != 200:
            raise TakeoffError(
                response.status_code,
                f"Tokenization failed\nStatus code: {str(response.status_code)}\nResponse: {response.text}",
            )
        if "tokens" not in response.json():
            raise TakeoffError(400, "Tokenization failed\nResponse: " + response.text)
        else:
            return response.json()["tokens"]

    def create_reader(self, reader_config: Dict[str, Any]) -> Dict[str, Any]:
        """Create a new reader.

        Args:
            reader_config (Dict[str, Any]): Dict containing all the reader configuration parameters.
        """
        try:
            reader = ReaderConfig(**reader_config)
        except Exception as e:
            raise TakeoffError(400, f"Reader creation failed\nError: {str(e)}")

        response = requests.post(self.mgmt_url + "/reader", json=reader.dict_without_optionals())
        if response.status_code != 201:
            raise TakeoffError(
                response.status_code,
                f"Reader creation failed\nStatus code: {str(response.status_code)}\nResponse: {response.text}",
            )
        return response.json()

    def delete_reader(self, reader_id: str) -> None:
        """Delete a reader, using their reader_id.

        Args:
            reader_id (str): Reader id.
        """
        response = requests.delete(self.mgmt_url + f"/reader/{reader_id}")
        if response.status_code != 204:
            raise TakeoffError(
                response.status_code,
                f"Reader deletion failed\nStatus code: {str(response.status_code)}\nResponse: {response.text}",
            )

    def list_all_readers(self) -> Dict[str, Dict[str, Any]]:
        """List all readers, ordering by consumer group.

        Returns:
            Dict[str, Dict[str, Any]]: List of reader ids.
        """
        response = requests.get(self.mgmt_url + "/reader_groups")
        if response.status_code != 200:
            raise TakeoffError(
                response.status_code,
                f"Reader listing failed\nStatus code: {str(response.status_code)}\nResponse: {response.text}",
            )
        return response.json()

    def get_reader_config(self, reader_id: str) -> Dict[str, Any]:
        """Get the config.json that a reader is running.

        Args:
            reader_id (str): Reader id.

        Returns:
            Dict[str, Any]: Reader configuration.
        """
        response = requests.get(self.mgmt_url + f"/config/{reader_id}")
        if response.status_code != 200:
            raise TakeoffError(
                response.status_code,
                f"Reader config retrieval failed\nStatus code: {str(response.status_code)}\nResponse: {response.text}",
            )
        return response.json()


if __name__ == "__main__":
    pass



class TakeoffError(Exception):
    """The takeoff error class"""

    def __init__(self, status_code: int, message: str, *args, **kwargs):
        """Create a new takeoff error

        Args:
            status_code (int): the http status code (from the takeoff server) that triggered the error
            message (str): the string message describing the error
        """
        super().__init__(message, *args, **kwargs)
        self.status_code = status_code
